@@ -2,30 +2,202 @@
 
 from __future__ import print_function
 
-import argparse
 import fnmatch
 import glob
 import json
 import os
 import sys
 import tempfile
+import time
+import types
 
 
 from functools import wraps
 from pprint import pprint
 
+import click
 import requests
 from requests import request
 
 
 GITHUB_API = "https://api.github.com"
 
+REQ_BUFFER_SIZE = 65536  # Chunk size when iterating a download body
+
+_github_token_cli_arg = None
+
+
+class _NoopProgressReporter(object):
+    reportProgress = False
+
+    def __init__(self, label='', length=0):
+        self.label = label
+        self.length = length
+
+    def update(self, chunk_size):
+        pass
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_value, tb):
+        pass
+
+
+progress_reporter_cls = _NoopProgressReporter
+"""The progress reporter class to instantiate. This class
+is expected to be a context manager with a constructor accepting `label`
+and `length` keyword arguments, an `update` method accepting a `chunk_size`
+argument and a class attribute `reportProgress` set to True (It can
+conveniently be initialized using `sys.stdout.isatty()`)
+"""
+
 
 def _request(*args, **kwargs):
     with_auth = kwargs.pop("with_auth", True)
-    if "GITHUB_TOKEN" in os.environ and with_auth:
-        kwargs["auth"] = (os.environ["GITHUB_TOKEN"], 'x-oauth-basic')
-    return request(*args, **kwargs)
+    token = _github_token_cli_arg
+    if not token:
+        token = os.environ.get("GITHUB_TOKEN", None)
+    if token and with_auth:
+        kwargs["auth"] = (token, 'x-oauth-basic')
+    for _ in range(3):
+        response = request(*args, **kwargs)
+        is_travis = os.getenv("TRAVIS",  None) is not None
+        if is_travis and 400 <= response.status_code < 500:
+            print("Retrying in 1s (%s Client Error: %s for url: %s)" % (
+                response.status_code, response.reason, response.url))
+            time.sleep(1)
+            continue
+        break
+    return response
+
+
+def handle_http_error(func):
+    @wraps(func)
+    def with_error_handling(*args, **kwargs):
+        try:
+            return func(*args, **kwargs)
+        except requests.exceptions.HTTPError as e:
+            print('Error sending {0} to {1}'.format(
+                e.request.method, e.request.url))
+            print('<', e.request.method, e.request.path_url)
+            for k in sorted(e.request.headers.keys()):
+                print('<', k, ':', e.request.headers[k])
+            if e.request.body:
+                print('<')
+                print('<', repr(e.request.body[:35]),
+                      '(total {0} bytes of data)'.format(len(e.request.body)))
+            print('')
+            print('>', e.response.status_code, e.response.reason)
+            for k in sorted(e.response.headers.keys()):
+                print('>', k.title(), ':', e.response.headers[k])
+            if e.response.content:
+                print('>')
+                print('>', repr(e.response.content[:35]),
+                      '(total {0} bytes of data)'.format(
+                          len(e.response.content)))
+            return 1
+    return with_error_handling
+
+
+def _check_for_credentials(func):
+    @wraps(func)
+    def with_check_for_credentials(*args, **kwargs):
+        has_github_token_env_var = "GITHUB_TOKEN" in os.environ
+        has_netrc = requests.utils.get_netrc_auth(GITHUB_API)
+        if (not _github_token_cli_arg
+                and not has_github_token_env_var and not has_netrc):
+            raise EnvironmentError(
+                "This command requires credentials provided by passing "
+                "--github-token CLI argument, set using GITHUB_TOKEN "
+                "env. variable or using netrc file. For more details, "
+                "see https://github.com/j0057/github-release#configuring")
+        return func(*args, **kwargs)
+    return with_check_for_credentials
+
+
+def _progress_bar(*args, **kwargs):
+    bar = click.progressbar(*args, **kwargs)
+    bar.bar_template = "  [%(bar)s]  %(info)s  %(label)s"
+    bar.show_percent = True
+    bar.show_pos = True
+
+    def formatSize(length):
+        if length == 0:
+            return '%.2f' % length
+        unit = ''
+        # See https://en.wikipedia.org/wiki/Binary_prefix
+        units = ['k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
+        while True:
+            if length <= 1024 or len(units) == 0:
+                break
+            unit = units.pop(0)
+            length /= 1024.
+        return '%.2f%s' % (length, unit)
+
+    def formatPos(_self):
+        pos = formatSize(_self.pos)
+        if _self.length_known:
+            pos += '/%s' % formatSize(_self.length)
+        return pos
+
+    bar.format_pos = types.MethodType(formatPos, bar)
+    return bar
+
+
+@click.group()
+@click.option("--github-token", envvar='GITHUB_TOKEN', default=None,
+              help="[default: GITHUB_TOKEN env. variable]")
+@click.option("--progress/--no-progress", default=True,
+              help="Display progress bar (default: yes).")
+def main(github_token, progress):
+    """A CLI to easily manage GitHub releases, assets and references."""
+    global progress_reporter_cls
+    progress_reporter_cls.reportProgress = sys.stdout.isatty() and progress
+    if progress_reporter_cls.reportProgress:
+        progress_reporter_cls = _progress_bar
+    global _github_token_cli_arg
+    _github_token_cli_arg = github_token
+
+
+@main.group("release")
+@click.argument('repo_name', metavar="REPOSITORY")
+@click.pass_context
+@handle_http_error
+def gh_release(ctx, repo_name):
+    """Manage releases (list, create, delete, ...) for
+    REPOSITORY (e.g jcfr/sandbox)
+    """
+    ctx.obj = repo_name
+
+
+# 1.6.0 (deprecated): Remove this bloc
+class AssetGroup(click.Group):
+    def get_command(self, ctx, cmd_name):
+        cmd_name = "delete" if cmd_name == "erase" else cmd_name
+        return click.Group.get_command(self, ctx, cmd_name)
+
+
+@main.group("asset", cls=AssetGroup)
+@click.argument('repo_name', metavar="REPOSITORY")
+@click.pass_context
+@handle_http_error
+def gh_asset(ctx, repo_name):
+    """Manage release assets (upload, download, ...) for
+    REPOSITORY (e.g jcfr/sandbox)
+    """
+    ctx.obj = repo_name
+
+
+@main.group("ref")
+@click.argument('repo_name', metavar="REPOSITORY")
+@click.pass_context
+@handle_http_error
+def gh_ref(ctx, repo_name):
+    """Manage references (list, create, delete, ...) for
+    REPOSITORY (e.g jcfr/sandbox)
+    """
+    ctx.obj = repo_name
 
 
 #
@@ -61,7 +233,7 @@ def get_releases(repo_name, verbose=False):
     releases = response.json()
     if verbose:
         list(map(print_release_info,
-                 sorted(response.json(), key=lambda r: r['tag_name'])))
+                 sorted(releases, key=lambda r: r['tag_name'])))
     return releases
 
 
@@ -103,15 +275,23 @@ def _update_release_sha(repo_name, tag_name, new_release_sha, dry_run):
     if previous_release_sha == new_release_sha:
         return
     tmp_tag_name = tag_name + "-tmp"
+
+    time.sleep(0.1)
     patch_release(repo_name, tag_name,
                   tag_name=tmp_tag_name,
                   target_commitish=new_release_sha,
                   dry_run=dry_run)
+
+    time.sleep(0.1)
     gh_ref_delete(repo_name, "refs/tags/%s" % tag_name, dry_run=dry_run)
+
+    time.sleep(0.1)
     patch_release(repo_name, tmp_tag_name,
                   tag_name=tag_name,
                   target_commitish=new_release_sha,
                   dry_run=dry_run)
+
+    time.sleep(0.1)
     gh_ref_delete(repo_name,
                   "refs/tags/%s" % tmp_tag_name, dry_run=dry_run)
 
@@ -164,39 +344,65 @@ def patch_release(repo_name, current_tag_name, **values):
             tags=True, verbose=verbose, dry_run=dry_run)
 
 
+def get_assets(repo_name, tag_name, verbose=False):
+    release = get_release(repo_name, tag_name)
+    if not release:
+        raise Exception('Release with tag_name {0} not found'.format(tag_name))
+    response = _request(
+        'GET', GITHUB_API + '/repos/{0}/releases/{1}/assets'.format(
+            repo_name, release["id"]))
+    response.raise_for_status()
+    assets = response.json()
+    if verbose:
+        for i, asset in enumerate(sorted(assets, key=lambda r: r['name'])):
+            print_asset_info(i, asset)
+    return assets
+
+
 def get_asset_info(repo_name, tag_name, filename):
-    release = get_release_info(repo_name, tag_name)
+    assets = get_assets(repo_name, tag_name)
     try:
-        asset = next(a for a in release['assets'] if a['name'] == filename)
+        asset = next(a for a in assets if a['name'] == filename)
         return asset
     except StopIteration:
         raise Exception('Asset with filename {0} not found in '
                         'release with tag_name {1}'.format(filename, tag_name))
 
 
-def gh_release_list(repo_name, verbose=True):
-    return get_releases(repo_name, verbose=verbose)
+@gh_release.command("list")
+@click.pass_obj
+def _cli_release_list(repo_name):
+    """List releases"""
+    return get_releases(repo_name, verbose=True)
 
 
-gh_release_list.description = {
-  "help": "List releases",
-  "params": ["repo_name"]
-}
-
-
-def gh_release_info(repo_name, tag_name):
+@gh_release.command("info")
+@click.argument("tag_name")
+@click.pass_obj
+def _cli_release_info(repo_name, tag_name):
+    """Get release description"""
     release = get_release_info(repo_name, tag_name)
     print_release_info(release)
 
 
-gh_release_info.description = {
-  "help": "Get release description",
-  "params": ["repo_name", "tag_name"]
-}
+@gh_release.command("create")
+@click.argument("tag_name")
+@click.argument("asset_pattern", nargs=-1)
+@click.option("--name")
+@click.option("--publish", is_flag=True, default=False)
+@click.option("--prerelease", is_flag=True, default=False)
+@click.option("--dry-run", is_flag=True, default=False)
+@click.option("--target-commitish")
+@click.pass_obj
+def cli_release_create(*args, **kwargs):
+    """Create a release"""
+    gh_release_create(*args, **kwargs)
 
 
-def gh_release_create(repo_name, tag_name, name=None,
-                      publish=False, prerelease=False, target_commitish=None):
+@_check_for_credentials
+def gh_release_create(repo_name, tag_name, asset_pattern=None, name=None,
+                      publish=False, prerelease=False,
+                      target_commitish=None, dry_run=False):
     if get_release(repo_name, tag_name) is not None:
         print('release %s: already exists\n' % tag_name)
         return
@@ -209,26 +415,37 @@ def gh_release_create(repo_name, tag_name, name=None,
         data["name"] = name
     if target_commitish is not None:
         data["target_commitish"] = target_commitish
-    response = _request(
-          'POST', GITHUB_API + '/repos/{0}/releases'.format(repo_name),
-          data=json.dumps(data),
-          headers={'Content-Type': 'application/json'})
-    response.raise_for_status()
-    print_release_info(response.json(), title="created '%s' release" % tag_name)
+    if not dry_run:
+        response = _request(
+              'POST', GITHUB_API + '/repos/{0}/releases'.format(repo_name),
+              data=json.dumps(data),
+              headers={'Content-Type': 'application/json'})
+        response.raise_for_status()
+        print_release_info(response.json(),
+                           title="created '%s' release" % tag_name)
+    else:
+        print("created '%s' release (dry_run)" % tag_name)
+    if asset_pattern is not None:
+        gh_asset_upload(repo_name, tag_name, asset_pattern, dry_run=dry_run)
 
 
-gh_release_create.description = {
-  "help": "Create a release",
-  "params": [
-      "repo_name", "tag_name", "name",
-      "publish", "prerelease", "target_commitish"
-  ],
-  "optional_params": {
-      "name": str, "publish": bool, "prerelease": bool, "target_commitish": str
-  }
-}
+@gh_release.command("edit")
+@click.argument("current_tag_name")
+@click.option("--tag-name", default=None)
+@click.option("--target-commitish", default=None)
+@click.option("--name", default=None)
+@click.option("--body", default=None)
+@click.option("--draft", is_flag=True, default=None)
+@click.option("--prerelease", is_flag=True, default=None)
+@click.option("--dry-run", is_flag=True, default=False)
+@click.option("--verbose", is_flag=True, default=False)
+@click.pass_obj
+def _cli_release_edit(*args, **kwargs):
+    """Edit a release"""
+    gh_release_edit(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_release_edit(repo_name, current_tag_name,
                     tag_name=None, target_commitish=None, name=None,
                     body=None,
@@ -243,26 +460,23 @@ def gh_release_edit(repo_name, current_tag_name,
     patch_release(repo_name, current_tag_name, **attributes)
 
 
-gh_release_edit.description = {
-  "help": "Edit a release",
-  "params": [
-      "repo_name", "current_tag_name",
-      "tag_name", "target_commitish", "name", "body", "draft", "prerelease",
-      "dry-run", "verbose"
-  ],
-  "optional_params": {
-      "tag_name": str, "target_commitish": str,
-      "name": str, "body": str, "draft": bool, "prerelease": bool,
-      "dry-run": bool, "verbose": bool
-  },
-  "optional_params_defaults": {"draft": None, "prerelease": None}
-}
+@gh_release.command("delete")
+@click.argument("pattern")
+@click.option("--keep-pattern")
+@click.option("--dry-run", is_flag=True, default=False)
+@click.option("--verbose", is_flag=True, default=False)
+@click.pass_obj
+def _cli_release_delete(*args, **kwargs):
+    """Delete selected release"""
+    gh_release_delete(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_release_delete(repo_name, pattern, keep_pattern=None,
                       dry_run=False, verbose=False):
     releases = get_releases(repo_name)
-    deleted = False
+    candidates = []
+    # Get list of candidate releases
     for release in releases:
         if not fnmatch.fnmatch(release['tag_name'], pattern):
             if verbose:
@@ -272,6 +486,9 @@ def gh_release_delete(repo_name, pattern, keep_pattern=None,
         if keep_pattern is not None:
             if fnmatch.fnmatch(release['tag_name'], keep_pattern):
                 continue
+        candidates.append(release['tag_name'])
+    for tag_name in candidates:
+        release = get_release(repo_name, tag_name)
         print('deleting release {0}'.format(release['tag_name']))
         if dry_run:
             continue
@@ -279,40 +496,47 @@ def gh_release_delete(repo_name, pattern, keep_pattern=None,
                + '/repos/{0}/releases/{1}'.format(repo_name, release['id']))
         response = _request('DELETE', url)
         response.raise_for_status()
-        deleted = True
-    return deleted
+    return len(candidates) > 0
 
 
-gh_release_delete.description = {
-  "help": "Delete selected releases",
-  "params": ["repo_name", "pattern", "keep_pattern", "dry-run", "verbose"],
-  "optional_params": {"keep_pattern": str, "dry-run": bool, "verbose": bool}
-}
+@gh_release.command("publish")
+@click.argument("tag_name")
+@click.option("--prerelease", is_flag=True, default=False)
+@click.pass_obj
+def _cli_release_publish(*args, **kwargs):
+    """Publish a release setting draft to 'False'"""
+    gh_release_publish(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_release_publish(repo_name, tag_name, prerelease=False):
     patch_release(repo_name, tag_name, draft=False, prerelease=prerelease)
 
 
-gh_release_publish.description = {
-  "help": "Publish a release setting draft to 'False'",
-  "params": ["repo_name", "tag_name", "prerelease"],
-  "optional_params": {"prerelease": bool}
-}
+@gh_release.command("unpublish")
+@click.argument("tag_name")
+@click.option("--prerelease", is_flag=True, default=False)
+@click.pass_obj
+def _cli_release_unpublish(*args, **kwargs):
+    """Unpublish a release setting draft to 'True'"""
+    gh_release_unpublish(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_release_unpublish(repo_name, tag_name, prerelease=False):
     draft = not prerelease
     patch_release(repo_name, tag_name, draft=draft, prerelease=prerelease)
 
 
-gh_release_unpublish.description = {
-  "help": "Unpublish a release setting draft to 'True'",
-  "params": ["repo_name", "tag_name", "prerelease"],
-  "optional_params": {"prerelease": bool}
-}
+@gh_release.command("release-notes")
+@click.argument("tag_name")
+@click.pass_obj
+def _cli_release_notes(*args, **kwargs):
+    """Set release notes"""
+    gh_release_notes(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_release_notes(repo_name, tag_name):
     release = get_release_info(repo_name, tag_name)
     (_, filename) = tempfile.mkstemp(suffix='.md')
@@ -333,23 +557,15 @@ def gh_release_notes(repo_name, tag_name):
         os.remove(filename)
 
 
-gh_release_notes.description = {
-  "help": "Set release notes",
-  "params": ["repo_name", "tag_name"]
-}
-
-
-def gh_release_debug(repo_name, tag_name):
+@gh_release.command("debug")
+@click.argument("tag_name")
+@click.pass_obj
+def _cli_release_debug(repo_name, tag_name):
+    """Print release detailed information"""
     release = get_release_info(repo_name, tag_name)
     pprint(release)
 
 
-gh_release_debug.description = {
-  "help": "Print release detailed information",
-  "params": ["repo_name", "tag_name"]
-}
-
-
 #
 # Assets
 #
@@ -358,89 +574,179 @@ def print_asset_info(i, asset, indent=""):
     print(indent + "Asset #{i}".format(i=i))
     indent = "  " + indent
     print(indent + "name      : {name}".format(i=i, **asset))
-    print(indent + "size      : {size}".format(i=i, **asset))
+    print(indent + "state     : {state}".format(i=i, **asset))
     print(indent + "uploader  : {login}".format(i=i, **asset['uploader']))
+    print(indent + "size      : {size}".format(i=i, **asset))
     print(indent + "URL       : {browser_download_url}".format(i=i, **asset))
     print(indent + "Downloads : {download_count}".format(i=i, **asset))
     print("")
 
 
-def gh_asset_upload(repo_name, tag_name, pattern, dry_run=False, verbose=False):
-    release = get_release_info(repo_name, tag_name)
-    uploaded = False
-    already_uploaded = False
-    upload_url = release["upload_url"]
-    if "{" in upload_url:
-        upload_url = upload_url[:upload_url.index("{")]
+@gh_asset.command("upload")
+@click.argument("tag_name")
+@click.argument("pattern", nargs=-1)
+@click.pass_obj
+def _cli_asset_upload(*args, **kwargs):
+    """Upload release assets"""
+    gh_asset_upload(*args, **kwargs)
 
-    if type(pattern) is list:
+
+class _ProgressFileReader(object):
+    """Wrapper used to capture File IO read progress."""
+    def __init__(self, stream, reporter):
+        self._stream = stream
+        self._reporter = reporter
+
+    def read(self, _size):
+        _chunk = self._stream.read(_size)
+        self._reporter.update(len(_chunk))
+        return _chunk
+
+    def __getattr__(self, attr):
+        return getattr(self._stream, attr)
+
+
+def _upload_release_file(
+        repo_name, tag_name, upload_url, filename,
+        verbose=False, dry_run=False, retry=True):
+    already_uploaded = False
+    uploaded = False
+    basename = os.path.basename(filename)
+    # Sanity checks
+    assets = get_assets(repo_name, tag_name)
+    download_url = None
+    for asset in assets:
+        if asset["name"] == basename:
+            if asset["state"] == "uploaded":
+                download_url = asset["browser_download_url"]
+                break
+            # Remove asset that failed to upload
+            # See https://developer.github.com/v3/repos/releases/#response-for-upstream-failure  # noqa: E501
+            if asset["state"] == "new":
+                print("  deleting %s (invalid asset "
+                      "with state set to 'new')" % asset['name'])
+                url = (
+                    GITHUB_API
+                    + '/repos/{0}/releases/assets/{1}'.format(
+                        repo_name, asset['id'])
+                )
+                response = _request('DELETE', url)
+                response.raise_for_status()
+
+    print("  uploading %s" % filename)
+
+    # Skip if an asset with same name has already been uploaded
+    # Trying to upload would give a HTTP error 422
+    if download_url:
+        already_uploaded = True
+        print("  skipping (asset with same name already exists)")
+        print("  download_url: %s" % download_url)
+        print("")
+        return already_uploaded, uploaded, {}
+    if dry_run:
+        uploaded = True
+        print("  download_url: Unknown (dry_run)")
+        print("")
+        return already_uploaded, uploaded, {}
+
+    url = '{0}?name={1}'.format(upload_url, basename)
+    if verbose and not progress_reporter_cls.reportProgress:
+        print("  upload_url: %s" % url)
+    file_size = os.path.getsize(filename)
+
+    # Attempt upload
+    with open(filename, 'rb') as f:
+        with progress_reporter_cls(
+                label=basename, length=file_size) as reporter:
+            response = _request(
+                'POST', url,
+                headers={'Content-Type': 'application/octet-stream'},
+                data=_ProgressFileReader(f, reporter))
+            data = response.json()
+
+    if response.status_code == 502 and retry:
+        print("  retrying (upload failed with status_code=502)")
+        already_uploaded, uploaded, data = _upload_release_file(
+            repo_name, tag_name, upload_url, filename,
+            verbose=verbose, retry=False)
+    else:
+        response.raise_for_status()
+    asset = data
+    print("  download_url: %s" % asset["browser_download_url"])
+    print("")
+    uploaded = True
+    return already_uploaded, uploaded, response.json()
+
+
+@_check_for_credentials
+def gh_asset_upload(repo_name, tag_name, pattern, dry_run=False, verbose=False):
+    if not dry_run:
+        upload_url = get_release_info(repo_name, tag_name)["upload_url"]
+        if "{" in upload_url:
+            upload_url = upload_url[:upload_url.index("{")]
+    else:
+        upload_url = "unknown"
+
+    # Raise exception if no token is specified AND netrc file is found
+    # BUT only api.github.com is specified. See #17
+    has_github_token = "GITHUB_TOKEN" in os.environ
+    has_netrc = requests.utils.get_netrc_auth(GITHUB_API)
+    if not has_github_token and has_netrc:
+        if requests.utils.get_netrc_auth(upload_url) is None:
+            raise EnvironmentError(
+                "Found netrc file but upload URL is missing. "
+                "For more details, "
+                "see https://github.com/j0057/github-release#configuring")
+
+    if type(pattern) in [list, tuple]:
         filenames = []
         for package in pattern:
             filenames.extend(glob.glob(package))
-        set(filenames)
-    else:
+        filenames = set(filenames)
+    elif pattern:
         filenames = glob.glob(pattern)
+    else:
+        filenames = []
 
     if len(filenames) > 0:
         print("uploading '%s' release asset(s) "
               "(found %s):" % (tag_name, len(filenames)))
 
+    uploaded = False
+    already_uploaded = False
+
     for filename in filenames:
-        print("  uploading %s" % filename)
-        basename = os.path.basename(filename)
-        # Skip if an asset with same name has already been uploaded
-        # Trying to upload would give a HTTP error 422
-        download_url = None
-        for asset in release["assets"]:
-            if asset["name"] == basename:
-                download_url = asset["browser_download_url"]
-                break
-        if download_url:
-            already_uploaded = True
-            print("  skipping (asset with same name already exists)")
-            print("  download_url: %s" % download_url)
-            print("")
-            continue
-        if dry_run:
-            uploaded = True
-            print("  download_url: Unknown (dry_run)")
-            print("")
-            continue
-        # Attempt upload
-        with open(filename, 'rb') as f:
-            url = '{0}?name={1}'.format(upload_url, basename)
-            if verbose:
-                print("  upload_url: %s" % url)
-            response = _request(
-                'POST', url,
-                headers={'Content-Type': 'application/octet-stream'},
-                data=f.read())
-            response.raise_for_status()
-            asset = response.json()
-            print("  download_url: %s" % asset["browser_download_url"])
-            print("")
-            uploaded = True
+        already_uploaded, uploaded, _ = _upload_release_file(
+            repo_name, tag_name, upload_url, filename, verbose, dry_run)
+
     if not uploaded and not already_uploaded:
         print("skipping upload of '%s' release assets ("
               "no files match pattern(s): %s)" % (tag_name, pattern))
         print("")
 
 
-gh_asset_upload.description = {
-  "help": "Upload release assets",
-  "params": ["repo_name", "tag_name", "pattern", "dry-run", "verbose"],
-  "optional_params": {"dry-run": bool, "verbose": bool}
-}
+@gh_asset.command("delete")
+@click.argument("tag_name")
+@click.argument("pattern")
+@click.option("--keep-pattern", default=None)
+@click.option("--dry-run", is_flag=True, default=False)
+@click.option("--verbose", is_flag=True, default=False)
+@click.pass_obj
+def _cli_asset_delete(*args, **kwargs):
+    """Delete selected release assets"""
+    gh_asset_delete(*args, **kwargs)
 
 
-def gh_asset_erase(repo_name, tag_name, pattern,
-                   keep_pattern=None, dry_run=False, verbose=False):
-    release = get_release_info(repo_name, tag_name)
+@_check_for_credentials
+def gh_asset_delete(repo_name, tag_name, pattern,
+                    keep_pattern=None, dry_run=False, verbose=False):
+    # Get assets
+    assets = get_assets(repo_name, tag_name)
     # List of assets
     excluded_assets = {}
     matched_assets = []
     matched_assets_to_keep = {}
-    for asset in release['assets']:
+    for asset in assets:
         if not fnmatch.fnmatch(asset['name'], pattern):
             skip_reason = "do NOT match pattern '%s'" % pattern
             excluded_assets[asset['name']] = skip_reason
@@ -485,14 +791,35 @@ def gh_asset_erase(repo_name, tag_name, pattern,
         print("")
 
 
-gh_asset_erase.description = {
-  "help": "Delete selected release assets",
-  "params": [
-      "repo_name", "tag_name", "pattern", "keep-pattern",
-      "dry-run", "verbose"
-  ],
-  "optional_params": {"keep-pattern": str, "dry-run": bool, "verbose": bool}
-}
+@gh_asset.command("download")
+@click.argument("tag_name")
+@click.argument("pattern", required=False)
+@click.pass_obj
+def _cli_asset_download(*args, **kwargs):
+    """Download release assets"""
+    gh_asset_download(*args, **kwargs)
+
+
+def _download_file(repo_name, asset):
+    response = _request(
+        method='GET',
+        url=GITHUB_API + '/repos/{0}/releases/assets/{1}'.format(
+            repo_name, asset['id']),
+        allow_redirects=False,
+        headers={'Accept': 'application/octet-stream'},
+        stream=True)
+    while response.status_code == 302:
+        response = _request(
+            'GET', response.headers['Location'], allow_redirects=False,
+            stream=True,
+            with_auth=False
+        )
+    with open(asset['name'], 'w+b') as f:
+        with progress_reporter_cls(
+                label=asset['name'], length=asset['size']) as reporter:
+            for chunk in response.iter_content(chunk_size=REQ_BUFFER_SIZE):
+                reporter.update(len(chunk))
+                f.write(chunk)
 
 
 def gh_asset_download(repo_name, tag_name=None, pattern=None):
@@ -513,28 +840,17 @@ def gh_asset_download(repo_name, tag_name=None, pattern=None):
                 continue
             print('release {0}: '
                   'downloading {1}'.format(release['tag_name'], asset['name']))
-            response = _request(
-                method='GET',
-                url=GITHUB_API + '/repos/{0}/releases/assets/{1}'.format(
-                    repo_name, asset['id']),
-                allow_redirects=False,
-                headers={'Accept': 'application/octet-stream'})
-            while response.status_code == 302:
-                response = _request(
-                    'GET', response.headers['Location'], allow_redirects=False,
-                    with_auth=False
-                )
-            with open(asset['name'], 'w+b') as f:
-                f.write(response.content)
+            _download_file(repo_name, asset)
             downloaded += 1
     return downloaded
 
 
-gh_asset_download.description = {
-  "help": "Download release assets",
-  "params": ["repo_name", "tag_name", "pattern"],
-  "optional_params": {"tag_name": str, "pattern": str}
-}
+@gh_asset.command("list")
+@click.argument("tag_name")
+@click.pass_obj
+def _cli_asset_list(repo_name, tag_name):
+    """List release assets"""
+    return get_assets(repo_name, tag_name, verbose=True)
 
 
 #
@@ -577,6 +893,16 @@ def get_refs(repo_name, tags=None, pattern=None):
     return filtered_data
 
 
+@gh_ref.command("list")
+@click.option("--tags", is_flag=True, default=False)
+@click.option("--pattern", default=None)
+@click.option("--verbose", is_flag=True, default=False)
+@click.pass_obj
+def _cli_ref_list(*args, **kwargs):
+    """List all references"""
+    gh_ref_list(*args, **kwargs)
+
+
 def gh_ref_list(repo_name, tags=None,  pattern=None, verbose=False):
     refs = get_refs(repo_name, tags=tags, pattern=pattern)
     sorted_refs = sorted(refs, key=lambda r: r['ref'])
@@ -586,13 +912,16 @@ def gh_ref_list(repo_name, tags=None,  pattern=None, verbose=False):
         list(map(lambda ref: print(ref['ref']), sorted_refs))
 
 
-gh_ref_list.description = {
-  "help": "List all references",
-  "params": ["repo_name", "tags", "pattern", "verbose"],
-  "optional_params": {"tags": bool, "pattern": str, "verbose": bool}
-}
+@gh_ref.command("create")
+@click.argument("reference")
+@click.argument("sha")
+@click.pass_obj
+def _cli_ref_create(*args, **kwargs):
+    """Create reference (e.g heads/foo, tags/foo)"""
+    gh_ref_create(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_ref_create(repo_name, reference, sha):
     data = {
         'ref': "refs/%s" % reference,
@@ -606,12 +935,19 @@ def gh_ref_create(repo_name, reference, sha):
     print_ref_info(response.json())
 
 
-gh_ref_create.description = {
-  "help": "Create reference (e.g heads/foo, tags/foo)",
-  "params": ["repo_name", "reference", "sha"]
-}
+@gh_ref.command("delete")
+@click.argument("pattern")
+@click.option("--keep-pattern", default=None)
+@click.option("--tags", is_flag=True, default=False)
+@click.option("--dry-run", is_flag=True, default=False)
+@click.option("--verbose", is_flag=True, default=False)
+@click.pass_obj
+def _cli_ref_delete(*args, **kwargs):
+    """Delete selected references"""
+    gh_ref_delete(*args, **kwargs)
 
 
+@_check_for_credentials
 def gh_ref_delete(repo_name, pattern, keep_pattern=None, tags=False,
                   dry_run=False, verbose=False):
     refs = get_refs(repo_name, tags=tags)
@@ -633,182 +969,6 @@ def gh_ref_delete(repo_name, pattern, keep_pattern=None, tags=False,
         response.raise_for_status()
 
 
-gh_ref_delete.description = {
-  "help": "Delete selected references",
-  "params": ["repo_name", "pattern", "keep_pattern", "tags",
-             "dry-run", "verbose"],
-  "optional_params": {"keep_pattern": str, "tags": bool,
-                      "dry-run": bool, "verbose": bool}
-}
-
-
-#
-# Decorators
-#
-
-def handle_http_error(func):
-    @wraps(func)
-    def with_error_handling(*args, **kwargs):
-        try:
-            return func(*args, **kwargs)
-        except requests.exceptions.HTTPError as e:
-            print('Error sending {0} to {1}'.format(
-                e.request.method, e.request.url))
-            print('<', e.request.method, e.request.path_url)
-            for k in sorted(e.request.headers.keys()):
-                print('<', k, ':', e.request.headers[k])
-            if e.request.body:
-                print('<')
-                print('<', repr(e.request.body[:35]),
-                      '(total {0} bytes of data)'.format(len(e.request.body)))
-            print('')
-            print('>', e.response.status_code, e.response.reason)
-            for k in sorted(e.response.headers.keys()):
-                print('>', k.title(), ':', e.response.headers[k])
-            if e.response.content:
-                print('>')
-                print('>', repr(e.response.content[:35]),
-                      '(total {0} bytes of data)'.format(
-                          len(e.response.content)))
-            return 1
-    return with_error_handling
-
-
-#
-# Command line parsing helpers
-#
-
-def _gh_parser(commands, prog=None):
-    parser = argparse.ArgumentParser(description=__doc__, prog=prog)
-    parser.add_argument(
-        "repo_name", type=str, metavar="REPOSITORY",
-        help="Github repository to update (e.g jcfr/sandbox)"
-    )
-    subparsers = parser.add_subparsers(help='sub-command help')
-
-    for command in commands:
-        func = commands[command]
-        cmd_help = func.description["help"]
-        cmd_params = list(func.description["params"])
-        cmd_opt_params = func.description.get("optional_params", {})
-        cmd_parser = subparsers.add_parser(command, help=cmd_help)
-        for cmd_param in cmd_params:
-            if cmd_param == "repo_name":  # parameter already specified above
-                continue
-            if cmd_param not in cmd_opt_params.keys():
-                cmd_parser.add_argument(cmd_param, type=str)
-            else:
-                if cmd_opt_params[cmd_param] is bool:
-                    cmd_parser.add_argument(
-                        "--%s" % cmd_param, action='store_true')
-                else:
-                    cmd_parser.add_argument(
-                        "--%s" % cmd_param, type=cmd_opt_params[cmd_param])
-        cmd_parser.set_defaults(func=func)
-
-        # Set defaults
-        params_defaults = func.description.get("optional_params_defaults", {})
-        cmd_parser.set_defaults(**params_defaults)
-
-    return parser
-
-
-def _gh_parse_arguments(commands, argv, prog):
-    args = _gh_parser(commands, prog).parse_args(argv)
-    func = args.func
-    return func(*[
-        vars(args).get(arg_name.replace("-", "_"), None)
-        for arg_name in func.description["params"]
-        ])
-
-
-#
-# Command line parsers
-#
-
-RELEASE_COMMANDS = {
-    'list': gh_release_list,
-    'info': gh_release_info,
-    'create': gh_release_create,
-    'edit': gh_release_edit,
-    'delete': gh_release_delete,
-    'publish': gh_release_publish,
-    'unpublish': gh_release_unpublish,
-    'release-notes': gh_release_notes,
-    'debug': gh_release_debug
-}
-
-
-@handle_http_error
-def gh_release(argv=None, prog=None):
-    return _gh_parse_arguments(RELEASE_COMMANDS, argv, prog)
-
-
-ASSET_COMMANDS = {
-    'upload': gh_asset_upload,
-    'download': gh_asset_download,
-    'delete': gh_asset_erase,
-    'erase': gh_asset_erase,
-}
-
-
-@handle_http_error
-def gh_asset(argv=None, prog=None):
-    return _gh_parse_arguments(ASSET_COMMANDS, argv, prog)
-
-
-REF_COMMANDS = {
-    'list': gh_ref_list,
-    'create': gh_ref_create,
-    'delete': gh_ref_delete
-}
-
-
-@handle_http_error
-def gh_ref(argv=None, prog=None):
-    return _gh_parse_arguments(REF_COMMANDS, argv, prog)
-
-
-PROG = os.path.basename(sys.argv[0])
-HELP = """Usage: {prog} COMMAND REPOSITORY [OPTIONS]
-       {prog} [-h]
-
-A CLI to easily manage GitHub releases, assets and references.
-
-Commands:
-    release    Manage releases (list, create, delete, ...)
-    asset      Manage release assets (upload, download, ...)
-    ref        Manage references (list, create, delete, ...)
-
-Repository:    Repository to update (e.g octocat/hello-worId)
-
-Options:
-    -h, --help       Show this help message and exit
-
-Run '{prog} COMMAND --help' for more information on a command.
-""".format(prog=PROG)
-
-
-def main():
-
-    if (len(sys.argv) == 1
-            or "--help" in sys.argv[:2]
-            or "-h" in sys.argv[:2]):
-        print(HELP)
-        exit(0)
-
-    parser = argparse.ArgumentParser()
-    parser.add_argument('command', default=None)
-
-    args = parser.parse_args(sys.argv[1:2])
-    if "gh_%s" % args.command not in globals():
-        print("%s: '%s' is not a githubrelease command." % (PROG, args.command))
-        print("See '%s --help'" % PROG)
-        exit(1)
-    globals()["gh_%s" % args.command](
-        sys.argv[2:], "%s %s" % (PROG, args.command))
-
-
 #
 # Script entry point
 #
